"""Tests of dataset opening options and driver choice"""
from contextlib import nullcontext
import concurrent.futures

import pytest

import rasterio
from rasterio.env import _GDAL_AT_LEAST_3_10, _GDAL_AT_LEAST_3_11
from rasterio.transform import Affine

def test_fail_with_missing_driver():
    """Fail to open a GeoTIFF without the GTiff driver"""
    with pytest.raises(rasterio.errors.RasterioIOError):
        rasterio.open('tests/data/RGB.byte.tif', driver='BMP')


def test_open_specific_driver():
    """Open a GeoTIFF with the GTiff driver"""
    with rasterio.open('tests/data/RGB.byte.tif', driver='GTiff') as src:
        assert src.count == 3


def test_open_specific_driver_with_options():
    """Open a GeoTIFF with the GTiff driver and GEOREF_SOURCES option"""
    with rasterio.open(
            'tests/data/RGB.byte.tif', driver='GTiff', GEOREF_SOURCES='NONE') as src:
        assert src.transform == Affine.identity()


def test_open_thread_safe(path_rgb_byte_tif, tmp_path):
    with (
        pytest.raises(rasterio.errors.GDALOptionNotImplementedError) if not _GDAL_AT_LEAST_3_10 else nullcontext(),
        rasterio.Env(GDAL_NUM_THREADS=2),
        rasterio.open(path_rgb_byte_tif, thread_safe=True, driver="LIBERTIFF" if _GDAL_AT_LEAST_3_11 else "GTiff") as src,
    ):
        def process(window):
            src.read(window=window).sum()

        windows = [window for ij, window in src.block_windows()]
        with concurrent.futures.ThreadPoolExecutor() as executor:
            executor.map(process, windows)



"""Classes capable of reading and writing datasets

Instances of these classes are called dataset objects.
"""

import logging
import warnings

from rasterio._base import get_dataset_driver, driver_can_create, driver_can_create_copy
from rasterio._io import (
    DatasetReaderBase,
    DatasetWriterBase,
    BufferedDatasetWriterBase,
    MemoryFileBase,
)
from rasterio.windows import WindowMethodsMixin
from rasterio.env import ensure_env
from rasterio.errors import RasterioDeprecationWarning
from rasterio.transform import TransformMethodsMixin
from rasterio._path import _UnparsedPath

try:
    from rasterio._filepath import FilePathBase
except ImportError:
    FilePathBase = object


log = logging.getLogger(__name__)


class DatasetReader(DatasetReaderBase, WindowMethodsMixin, TransformMethodsMixin):
    """An unbuffered data and metadata reader"""

    def __repr__(self):
        return "<{} DatasetReader name='{}' mode='{}'>".format(
            self.closed and 'closed' or 'open', self.name, self.mode)


class DatasetWriter(DatasetWriterBase, WindowMethodsMixin, TransformMethodsMixin):
    """An unbuffered data and metadata writer. Its methods write data
    directly to disk.
    """

    def __repr__(self):
        return "<{} DatasetWriter name='{}' mode='{}'>".format(
            self.closed and 'closed' or 'open', self.name, self.mode)


class BufferedDatasetWriter(
    BufferedDatasetWriterBase, WindowMethodsMixin, TransformMethodsMixin
):
    """Maintains data and metadata in a buffer, writing to disk or
    network only when `close()` is called.

    This allows incremental updates to datasets using formats that don't
    otherwise support updates, such as JPEG.
    """

    def __repr__(self):
        return "<{} BufferedDatasetWriter name='{}' mode='{}'>".format(
            self.closed and 'closed' or 'open', self.name, self.mode)


class MemoryFile(MemoryFileBase):
    """A BytesIO-like object, backed by an in-memory file.

    This allows formatted files to be read and written without I/O.

    A MemoryFile created with initial bytes becomes immutable. A
    MemoryFile created without initial bytes may be written to using
    either file-like or dataset interfaces.

    Examples
    --------

    A GeoTIFF can be loaded in memory and accessed using the GeoTIFF
    format driver

    >>> with open('tests/data/RGB.byte.tif', 'rb') as f, MemoryFile(f) as memfile:
    ...     with memfile.open() as src:
    ...         pprint.pprint(src.profile)
    ...
    {'count': 3,
     'crs': CRS({'init': 'epsg:32618'}),
     'driver': 'GTiff',
     'dtype': 'uint8',
     'height': 718,
     'interleave': 'pixel',
     'nodata': 0.0,
     'tiled': False,
     'transform': Affine(300.0379266750948, 0.0, 101985.0,
           0.0, -300.041782729805, 2826915.0),
     'width': 791}

    """

    def __init__(self, file_or_bytes=None, dirname=None, filename=None, ext=".tif"):
        """Create a new file in memory

        Parameters
        ----------
        file_or_bytes : file-like object or bytes, optional
            File or bytes holding initial data.
        filename : str, optional
            An optional filename. A unique one will otherwise be generated.
        ext : str, optional
            An optional extension.

        Returns
        -------
        MemoryFile
        """
        super().__init__(
            file_or_bytes=file_or_bytes, dirname=dirname, filename=filename, ext=ext
        )

    @ensure_env
    def open(self, driver=None, width=None, height=None, count=None, crs=None,
             transform=None, dtype=None, nodata=None, sharing=False, thread_safe=False, **kwargs):
        """Open the file and return a Rasterio dataset object.

        If data has already been written, the file is opened in 'r'
        mode. Otherwise, the file is opened in 'w' mode.

        Parameters
        ----------
        Note well that there is no `path` parameter: a `MemoryFile`
        contains a single dataset and there is no need to specify a
        path.

        Other parameters are optional and have the same semantics as the
        parameters of `rasterio.open()`.
        """
        mempath = _UnparsedPath(self.name)

        if self.closed:
            raise ValueError("I/O operation on closed file.")
        if len(self) > 0:
            log.debug(f"VSI path: {mempath.path}")
            rd = DatasetReader(mempath, driver=driver, sharing=sharing, thread_safe=thread_safe, **kwargs)
        else:
            writer = get_writer_for_driver(driver)
            rd = writer(
                mempath,
                "w+",
                driver=driver,
                width=width,
                height=height,
                count=count,
                crs=crs,
                transform=transform,
                dtype=dtype,
                nodata=nodata,
                sharing=sharing,
                **kwargs
            )

        # Push the new dataset's context exit onto the MemoryFile's ExitStack.
        # This ensures that when this MemoryFile is closed, any derived dataset
        # is also closed automatically.
        self._env.push(rd)
        return rd

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if self._env:
            self._env.close()
        self.close()


class _FilePath(FilePathBase):
    """A BytesIO-like object, backed by a Python file object.

    .. deprecated:: 1.4.0
       FilePath is supplanted by open's new opener keyword argument,
       and will be removed in 2.0.0.

    """
    def __init__(self, filelike_obj, dirname=None, filename=None):
        """Create a new wrapper around the provided file-like object.

        Parameters
        ----------
        filelike_obj : file-like object
            Open file-like object. Currently only reading is supported.
        filename : str, optional
            An optional filename. A unique one will otherwise be generated.

        Returns
        -------
        PythonVSIFile
        """
        warnings.warn(
            "FilePath is supplanted by open's new opener keyword argument, and will be removed in 2.0.0.",
            RasterioDeprecationWarning,
        )
        super().__init__(
            filelike_obj, dirname=dirname, filename=filename
        )

    @ensure_env
    def open(self, driver=None, sharing=False, thread_safe=False, **kwargs):
        """Open the file and return a Rasterio dataset object.

        The provided file-like object is assumed to be readable.
        Writing is currently not supported.

        Parameters are optional and have the same semantics as the
        parameters of `rasterio.open()`.

        Returns
        -------
        DatasetReader

        Raises
        ------
        IOError
            If the memory file is closed.

        """
        mempath = _UnparsedPath(self.name)

        if self.closed:
            raise OSError("I/O operation on closed file.")

        # Assume we were given a non-empty file-like object
        log.debug(f"VSI path: {mempath.path}")

        return DatasetReader(mempath, driver=driver, sharing=sharing, thread_safe=thread_safe, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


if FilePathBase is not object:
    # only make this object available if the cython extension was compiled
    FilePath = _FilePath


class ZipMemoryFile(MemoryFile):
    """A read-only BytesIO-like object backed by an in-memory zip file.

    This allows a zip file containing formatted files to be read
    without I/O.
    """

    def __init__(self, file_or_bytes=None):
        super().__init__(file_or_bytes, ext="zip")

    @ensure_env
    def open(self, path, driver=None, sharing=False, thread_safe=False, **kwargs):
        """Open a dataset within the zipped stream.

        Parameters
        ----------
        path : str
            Path to a dataset in the zip file, relative to the root of the
            archive.

        Other parameters are optional and have the same semantics as the
        parameters of `rasterio.open()`.

        Returns
        -------
        A Rasterio dataset object
        """
        zippath = _UnparsedPath("/vsizip{}/{}".format(self.name, path.lstrip("/")))

        if self.closed:
            raise ValueError("I/O operation on closed file.")
        return DatasetReader(zippath, driver=driver, sharing=sharing, thread_safe=thread_safe, **kwargs)


def get_writer_for_driver(driver):
    """Return the writer class appropriate for the specified driver."""
    if not driver:
        raise ValueError("'driver' is required to read/write dataset.")
    cls = None
    if driver_can_create(driver):
        cls = DatasetWriter
    elif driver_can_create_copy(driver):  # pragma: no branch
        cls = BufferedDatasetWriter
    return cls


def get_writer_for_path(path, driver=None):
    """Return the writer class appropriate for the existing dataset."""
    if not driver:
        driver = get_dataset_driver(path)
    return get_writer_for_driver(driver)



"""
Registry of common rio CLI options.  See cligj for more options.

.. code-block:: none

    -a, --all: Use all pixels touched by features.  In rio-mask, rio-rasterize
    --as-mask/--not-as-mask: interpret band as mask or not.  In rio-shapes
    --band/--mask: use band or mask.  In rio-shapes
    --bbox:
    -b, --bidx: band index(es) (singular or multiple value versions).
        In rio-info, rio-sample, rio-shapes, rio-stack (different usages)
    --bounds: bounds in world coordinates.
        In rio-info, rio-rasterize (different usages)
    --count: count of bands.  In rio-info
    --crop: Crop raster to extent of features.  In rio-mask
    --crs: CRS of input raster.  In rio-info
    --default-value: default for rasterized pixels.  In rio-rasterize
    --dimensions: Output width, height.  In rio-rasterize
    --dst-crs: destination CRS.  In rio-transform
    --fill: fill value for pixels not covered by features.  In rio-rasterize
    --formats: list available formats.  In rio-info
    --height: height of raster.  In rio-info
    -i, --invert: Invert mask created from features: In rio-mask
    -j, --geojson-mask: GeoJSON for masking raster.  In rio-mask
    --lnglat: geographic coordinates of center of raster.  In rio-info
    --masked/--not-masked: read masked data from source file.
        In rio-calc, rio-info
    -m, --mode: output file mode (r, r+).  In rio-insp
    --name: input file name alias.  In rio-calc
    --nodata: nodata value.  In rio-info, rio-merge (different usages)
    --photometric: photometric interpretation.  In rio-stack
    --property: GeoJSON property to use as values for rasterize.  In rio-rasterize
    -r, --res: output resolution.
        In rio-info, rio-rasterize (different usages.  TODO: try to combine
        usages, prefer rio-rasterize version)
    --sampling: Inverse of sampling fraction.  In rio-shapes
    --shape: shape (width, height) of band.  In rio-info
    --src-crs: source CRS.
        In rio-insp, rio-rasterize (different usages.  TODO: consolidate usages)
    --stats: print raster stats.  In rio-inf
    -t, --dtype: data type.  In rio-calc, rio-info (different usages)
    --width: width of raster.  In rio-info
    --with-nodata/--without-nodata: include nodata regions or not.  In rio-shapes.
    -v, --tell-me-more, --verbose
    --vfs: virtual file system.
"""

import logging
import os
import re

import click

import rasterio
import rasterio.shutil
from rasterio.dtypes import (
    ubyte,
    uint8,
    uint16,
    uint32,
    uint64,
    int8,
    int16,
    int32,
    int64,
    float16,
    float32,
    float64
)
from rasterio._path import _parse_path, _UnparsedPath


logger = logging.getLogger(__name__)


def _cb_key_val(ctx, param, value):

    """
    click callback to validate `--opt KEY1=VAL1 --opt KEY2=VAL2` and collect
    in a dictionary like the one below, which is what the CLI function receives.
    If no value or `None` is received then an empty dictionary is returned.

        {
            'KEY1': 'VAL1',
            'KEY2': 'VAL2'
        }

    Note: `==VAL` breaks this as `str.split('=', 1)` is used.
    """

    if not value:
        return {}
    else:
        out = {}
        for pair in value:
            if "=" not in pair:
                raise click.BadParameter(f"Invalid syntax for KEY=VAL arg: {pair}")
            else:
                k, v = pair.split('=', 1)
                k = k.lower()
                v = v.lower()
                out[k] = None if v.lower() in ['none', 'null', 'nil', 'nada'] else v
        return out


def abspath_forward_slashes(path):
    """Return forward-slashed version of os.path.abspath"""
    return '/'.join(os.path.abspath(path).split(os.path.sep))


def file_in_handler(ctx, param, value):
    """Normalize ordinary filesystem and VFS paths"""
    try:
        path = _parse_path(value)

        if isinstance(path, _UnparsedPath):

            if os.path.exists(path.path) and rasterio.shutil.exists(value):
                return abspath_forward_slashes(path.path)
            else:
                return path.name

        elif path.scheme and path.is_remote:
            return path.name

        elif path.archive:
            if os.path.exists(path.archive) and rasterio.shutil.exists(value):
                archive = abspath_forward_slashes(path.archive)
                return f"{path.scheme}://{archive}!{path.path}"
            else:
                raise OSError(f"Input archive {path.archive} does not exist")

        else:
            if os.path.exists(path.path) and rasterio.shutil.exists(value):
                return abspath_forward_slashes(path.path)
            else:
                raise OSError(f"Input file {path.path} does not exist")

    except Exception:
        raise click.BadParameter(f"{value} is not a valid input file")


def files_in_handler(ctx, param, value):
    """Process and validate input file names"""
    return tuple(file_in_handler(ctx, param, item) for item in value)


def files_inout_handler(ctx, param, value):
    """Process and validate input file names"""
    return tuple(file_in_handler(ctx, param, item) for item in value[:-1]) + tuple(value[-1:])


def from_like_context(ctx, param, value):
    """Return the value for an option from the context if the option
    or `--all` is given, else return None."""
    if ctx.obj and ctx.obj.get('like') and (
            value == 'like' or ctx.obj.get('all_like')):
        return ctx.obj['like'][param.name]
    else:
        return None


def like_handler(ctx, param, value):
    """Copy a dataset's meta property to the command context for access
    from other callbacks."""
    if ctx.obj is None:
        ctx.obj = {}
    if value:
        with rasterio.open(value) as src:
            metadata = src.meta
            ctx.obj['like'] = metadata
            ctx.obj['like']['transform'] = metadata['transform']
            ctx.obj['like']['tags'] = src.tags()
            ctx.obj['like']['colorinterp'] = src.colorinterp
        return True
    else:  # pragma: no cover
        return False


def nodata_handler(ctx, param, value):
    """Return a float or None"""
    if value is None or value.lower() in ["null", "nil", "none", "nada"]:
        return None
    else:
        try:
            return float(value)
        except (TypeError, ValueError):
            raise click.BadParameter(
                f"{value!r} is not a number", param=param, param_hint="nodata"
            )


def edit_nodata_handler(ctx, param, value):
    """Get nodata value from a template file or command line.

    Expected values are 'like', 'null', a numeric value, 'nan', or None.

    Returns
    -------
    float or None

    Raises
    ------
    click.BadParameter

    """
    if value == "like" or value is None:
        retval = from_like_context(ctx, param, value)
        if retval is not None:
            return retval

    return nodata_handler(ctx, param, value)


def bounds_handler(ctx, param, value):
    """Handle different forms of bounds."""
    retval = from_like_context(ctx, param, value)
    if retval is None and value is not None:
        try:
            value = value.strip(', []')
            retval = tuple(float(x) for x in re.split(r'[,\s]+', value))
            assert len(retval) == 4
            return retval
        except Exception:
            raise click.BadParameter(
                "{!r} is not a valid bounding box representation".format(value)
            )
    else:  # pragma: no cover
        return retval


# Singular input file
file_in_arg = click.argument('INPUT', callback=file_in_handler)

# Singular output file
file_out_arg = click.argument("OUTPUT", type=click.Path())

# Multiple input files.
files_in_arg = click.argument(
    'files',
    nargs=-1,
    type=click.Path(),
    required=True,
    metavar="INPUTS...",
    callback=files_in_handler)

# Multiple files, last of which is an output file.
files_inout_arg = click.argument(
    'files',
    nargs=-1,
    type=click.Path(),
    required=True,
    metavar="INPUTS... OUTPUT",
    callback=files_inout_handler)

bidx_opt = click.option(
    '-b', '--bidx',
    type=int,
    default=1,
    help="Input file band index (default: 1).")

bidx_mult_opt = click.option(
    '-b', '--bidx',
    type=int,
    multiple=True,
    help="Indexes of input file bands.")

# The 'magic' --bidx option is especially for rio-stack, which supports
# selection of multiple bands from multiple files. It has 'str' type
# instead of 'int' to support ranges in a syntax like `--bidx 1..3`.
bidx_magic_opt = click.option(
    '-b', '--bidx',
    multiple=True,
    help="Indexes of input file bands.")

# TODO: may be better suited to cligj?
bounds_opt = click.option(
    "--bounds",
    default=None,
    callback=bounds_handler,
    help="Bounds: 'left bottom right top' or '[left, bottom, right, top]'.",
)

dimensions_opt = click.option(
    '--dimensions',
    nargs=2, type=int, default=None,
    help='Output dataset width, height in number of pixels.')

dtype_opt = click.option(
    '-t', '--dtype',
    type=click.Choice([
        ubyte,
        uint8,
        uint16,
        uint32,
        uint64,
        int8,
        int16,
        uint32,
        int32,
        int64,
        float16,
        float32,
        float64
    ]),
    default=None,
    help="Output data type.")

like_file_opt = click.option(
    '--like',
    type=click.Path(exists=True),
    help='Raster dataset to use as a template for obtaining affine '
         'transform (bounds and resolution), crs, data type, and driver '
         'used to create the output.')

masked_opt = click.option(
    '--masked/--not-masked',
    default=True,
    help="Evaluate expressions using masked arrays (the default) or ordinary "
         "numpy arrays.")

output_opt = click.option(
    '-o', '--output',
    default=None,
    type=click.Path(),
    help="Path to output file (optional alternative to a positional arg).")

resolution_opt = click.option(
    '-r', '--res',
    multiple=True, type=float, default=None,
    help='Output dataset resolution in units of coordinate '
         'reference system. Pixels assumed to be square if this option '
         'is used once, otherwise use: '
         '--res pixel_width --res pixel_height.')

creation_options = click.option(
    '--co', '--profile', 'creation_options',
    metavar='NAME=VALUE',
    multiple=True,
    callback=_cb_key_val,
    help="Driver specific creation options. "
         "See the documentation for the selected output driver for "
         "more information.")

rgb_opt = click.option(
    '--rgb', 'photometric',
    flag_value='rgb',
    help="Set RGB photometric interpretation.")

overwrite_opt = click.option(
    '--overwrite', 'overwrite',
    is_flag=True, type=bool, default=False,
    help="Always overwrite an existing output file.")

nodata_opt = click.option(
    "--nodata",
    callback=nodata_handler,
    default=None,
    metavar="NUMBER|nan",
    help="Set a Nodata value.",
)

edit_nodata_opt = click.option(
    "--nodata",
    callback=edit_nodata_handler,
    default=None,
    metavar="NUMBER|nan|null",
    help="Modify the Nodata value.",
)

like_opt = click.option(
    '--like',
    type=click.Path(exists=True),
    callback=like_handler,
    is_eager=True,
    help="Raster dataset to use as a template for obtaining affine "
         "transform (bounds and resolution), crs, and nodata values.")

all_touched_opt = click.option(
    '-a', '--all', '--all_touched', 'all_touched',
    is_flag=True,
    default=False,
    help='Use all pixels touched by features, otherwise (default) use only '
         'pixels whose center is within the polygon or that are selected by '
         'Bresenhams line algorithm')

# Feature collection or feature sequence switch.
sequence_opt = click.option(
    '--sequence/--collection',
    default=True,
    help="Write a LF-delimited sequence of texts containing individual "
         "objects (the default) or write a single JSON text containing a "
         "feature collection object.")

format_opt = click.option(
    "-f", "--format", "--driver", "driver", help="Output format driver."
)


def geojson_type_opt(*, allowed, default):
    """GeoJSON output mode"""
    def verify_geojson_type(ctx, param, value):
        geojson_type = ctx.params.setdefault("geojson_type", {})
        geojson_type[param.name] = value
        if len(geojson_type) == len(allowed):
            # We've now seen all options for geojson_type, so verify their contents.
            specified = [
                name.removeprefix("geojson_type_")
                for name, value in geojson_type.items()
                if value
            ]
            if len(specified) > 1:
                names = ",".join(f"--{name}" for name in specified)
                raise click.BadParameter(
                    f"too many GeoJSON output formats: {names} may not be specified together"
                )
            ctx.params["geojson_type"] = specified[0] if specified else default

    def wrapper(func):
        options = {
            "collection": "feature collection(s)",
            "feature": "feature(s)",
            "bbox": "bounding box array(s)",
        }

        for name in allowed:
            if name not in options:
                raise click.UsageError(f"{name} is not a valid GeoJSON output mode")
            help_suffix = " (default)" if name == default else ""
            opt = click.option(
                f"--{name}",
                f"geojson_type_{name}",
                is_flag=True,
                expose_value=False,
                callback=verify_geojson_type,
                help=f"Output as GeoJSON {options[name]}{help_suffix}",
            )
            func = opt(func)
        return func

    return wrapper



#!/usr/bin/env python

# These following environmental variables influence this script:
#
# GDAL_CONFIG: the path to a gdal-config program that points to GDAL headers,
# libraries, and data files.
#
# PACKAGE_DATA: if defined, GDAL and PROJ4 data files will be copied into the
# source or binary distribution. This is essential when creating self-contained
# binary wheels.
#
# PATH: if we don't find `gdal-config`, we instead search for the `gdalinfo`
# executable using the `PATH` environment variable. We run it to get the GDAL
# version, and we compute the include, library, share, etc. directory paths
# from that executable's path.

import copy
import logging
import os
import platform
import pprint
import re
import shutil
import sys
from subprocess import check_output

from setuptools import setup
from setuptools.extension import Extension

logging.basicConfig(stream=sys.stderr, level=logging.INFO)
log = logging.getLogger()


def copy_data_tree(datadir, destdir):
    try:
        shutil.rmtree(destdir)
    except OSError:
        pass
    shutil.copytree(datadir, destdir)


# python -W all setup.py ...
if "all" in sys.warnoptions:
    log.level = logging.DEBUG

# Use Cython if available.
try:
    from Cython.Build import cythonize
except ImportError:
    raise SystemExit(
        "ERROR: Cython.Build.cythonize not found. "
        "Cython is required to build rasterio.")

# By default we'll try to get options via gdal-config. On systems without,
# options will need to be set in setup.cfg, on the setup command line, or through
# the GDAL_INSTALL_PREFIX environment variable.
include_dirs: list[str] = []
library_dirs: list[str] = []
libraries: list[str] = []
extra_link_args: list[str] = []
gdal2plus: bool = False
gdal_output: list[str | None] = [None] * 4
gdalversion: str | None = None
gdal_major_version: int = 0
gdal_minor_version: int = 0
gdal_patch_version: int = 0
gdal_data_dir: str | None = None

try:
    import numpy as np

    include_dirs.append(np.get_include())
except ImportError:
    raise SystemExit("ERROR: Numpy and its headers are required to run setup().")


def _find_executable(executable_name: str) -> str | None:
    """Find the given executable in PATH.

    Return the absolute path to the executable if found, otherwise None.
    """

    log.info("Searching for executable %s on sys.prefix")
    executable_path = shutil.which(executable_name, path=sys.prefix)
    if executable_path is None:
        log.info("Did not find executable %s on sys.prefix, searching on PATH")
        executable_path = shutil.which(executable_name, path=os.environ["PATH"])
    if executable_path is None:
        log.info("Did not find executable %s on sys.prefix or on PATH", executable_name)
        return None
    executable_path = os.path.abspath(executable_path)
    log.info("Found executable %s at path %s", executable_name, executable_path)
    return executable_path


def find_gdal_install_with_executable(
    executable_name: str = "gdalinfo",
) -> tuple[str, tuple[int, int, int]] | None:
    """Find the given GDAL executable and get the GDAL version and install directory from it.

    We first search on `sys.prefix`, then on `os.environ["PATH"]`.

    executable_name should be the name of a GDAL executable, like `"gdalinfo"`, with or without extension.
    It should be executable on the command line through this name.

    Return the absolute path to GDAL's install path (above bin, include, share, and lib) and the major, minor, and patch
    numbers of GDAL's version, if the executable was found. Otherwise, return None.

    This does not check if the install is complete.
    """

    executable_path = _find_executable(executable_name)
    # Run it to get the GDAL version
    result = check_output([executable_name, "--version"], text=True).strip()
    version_match = re.match(
        r"^GDAL (\d+)\.(\d+)\.(\d+) .*$", result, flags=re.MULTILINE
    )
    if version_match:
        gdal_version: tuple[int, int, int] = (
            int(version_match.group(1)),
            int(version_match.group(2)),
            int(version_match.group(3)),
        )
    else:
        raise RuntimeError(
            "Could not parse GDAL version from output of `%s --version`",
            executable_path,
        )
    # Move up two levels: parent of the `bin` folder
    gdal_install_prefix = os.path.dirname(os.path.dirname(executable_path))
    return gdal_install_prefix, gdal_version

def fill_gdal_build_options_using_executable(executable_name: str) -> None:
    """Fill the global GDAL build options using information from the provided install path.

    executable_name should be the name of a GDAL executable, like `"gdalinfo"`, with or without extension.
    It should be executable on the command line through this name.

    Raise `RuntimeError` if we don't find a GDAL install.
    """

    global \
        gdal_data_dir, \
        gdalversion, \
        gdal_major_version, \
        gdal_minor_version, \
        gdal_patch_version

    prefix_search_result = find_gdal_install_with_executable(
        executable_name=executable_name
    )
    if prefix_search_result is None:
        raise RuntimeError(
            'Could not find the GDAL install directory using executable "%s"',
            executable_name,
        )

    gdal_install_prefix, gdal_version_numbers = prefix_search_result

    # Find library directory
    found_library_dir = os.path.join(gdal_install_prefix, "lib")
    if not os.path.isdir(found_library_dir):
        raise RuntimeError("Could not find library directory at %s", found_library_dir)

    # Does the GDAL library file exist?
    found_gdal_lib = os.path.join(
        found_library_dir, "gdal.lib" if os.name == "nt" else "gdal.so"
    )
    if not os.path.isfile(found_gdal_lib):
        raise RuntimeError("Could not find GDAL library file at %s", found_library_dir)
    found_gdal_lib_name = os.path.splitext(os.path.basename(found_gdal_lib))[0]

    # Find include directory
    found_include_dir = os.path.join(gdal_install_prefix, "include")
    if not os.path.isdir(found_include_dir):
        raise RuntimeError("Could not find include directory at %s", found_include_dir)
    # Check that at least one header exists
    found_gdal_header = os.path.join(found_include_dir, "gdal.h")
    if not os.path.isfile(found_gdal_header):
        raise RuntimeError("Could not find GDAL header at %s", found_gdal_header)

    # Find GDAL_DATA directory
    found_gdal_data_dir = os.path.join(gdal_install_prefix, "share", "gdal")
    if not os.path.isdir(found_gdal_data_dir):
        raise RuntimeError(
            "Could not find GDAL_DATA directory at %s", found_gdal_data_dir
        )

    log.info(
        "Found a GDAL install at %s that seems complete, using it", gdal_install_prefix
    )
    library_dirs.append(found_library_dir)
    include_dirs.append(found_include_dir)
    gdalversion = ".".join(map(str, gdal_version_numbers))
    gdal_major_version, gdal_minor_version, gdal_patch_version = gdal_version_numbers
    libraries.append(found_gdal_lib_name)
    gdal_data_dir = found_gdal_data_dir

def fill_gdal_build_options_using_gdal_config() -> None:
    """Run gdal-config and fill the global build options using its output.

    If gdal-config's path isn't found, raise `FileNotFoundError`.
    If it fails, raise `CalledProcessError`.
    """

    global gdalversion

    log.info("Using gdal-config to get GDAL build options")
    gdal_config = os.environ.get("GDAL_CONFIG", _find_executable("gdal-config"))
    for i, flag in enumerate(("--cflags", "--libs", "--datadir", "--version")):
        try:
            gdal_output[i] = check_output([gdal_config, flag]).decode("utf-8").strip()
        except FileNotFoundError as error:
            raise FileNotFoundError(
                'gdal-config not found under the name "%s"', gdal_config
            ) from error

    for item in gdal_output[0].split():
        if item.startswith("-I"):
            include_dirs.extend(item[2:].split(":"))
    for item in gdal_output[1].split():
        if item.startswith("-L"):
            library_dirs.extend(item[2:].split(":"))
        elif item.startswith("-l"):
            libraries.append(item[2:])
        else:
            # e.g. -framework GDAL
            extra_link_args.append(item)
    # datadir, gdal_output[2] handled below

    gdalversion = gdal_output[3]
    if gdalversion:
        log.info("GDAL API version obtained from gdal-config: %s", gdalversion)

if "clean" not in sys.argv:
    try:
        fill_gdal_build_options_using_gdal_config()
    except Exception as e:
        # Try to run gdalinfo and get information from that instead
        log.info(
            "Failed to use gdal-config, trying to run gdalinfo instead (gdal-config error of type %s: %s)",
            type(e).__name__,
            e,
        )
        try:
            fill_gdal_build_options_using_executable(executable_name="gdalinfo")
        except Exception as e:
            log.warning(
                "Failed to get options via both gdal-config and gdalinfo. (gdalinfo error of type %s: %s)",
                type(e).__name__,
                e,
            )

    # Get GDAL API version from environment variable.
    if 'GDAL_VERSION' in os.environ:
        gdalversion = os.environ['GDAL_VERSION']
        log.info("GDAL API version obtained from environment: %s", gdalversion)

    # Get GDAL API version from the command line if specified there.
    if '--gdalversion' in sys.argv:
        index = sys.argv.index('--gdalversion')
        sys.argv.pop(index)
        gdalversion = sys.argv.pop(index)
        log.info("GDAL API version obtained from command line option: %s",
                 gdalversion)

    if not gdalversion:
        raise SystemExit("ERROR: A GDAL API version must be specified. Provide a path "
                 "to gdal-config using a GDAL_CONFIG environment variable "
                 "or use a GDAL_VERSION environment variable.")

    gdal_major_version, gdal_minor_version, gdal_patch_version = map(
        int, re.findall("[0-9]+", gdalversion)[:3]
    )

    if (gdal_major_version, gdal_minor_version) < (3, 8):
        raise SystemExit("ERROR: GDAL >= 3.8 is required for rasterio. "
                 "Please upgrade GDAL.")

# Conditionally copy the GDAL data. To be used in conjunction with
# the bdist_wheel command to make self-contained binary wheels.
if os.environ.get("PACKAGE_DATA"):
    gdal_data_destination = "rasterio/gdal_data"
    if gdal_data_dir is None:
        if gdal_output[2]:
           gdal_data_dir = gdal_output[2]
        else:
            # check to see if GDAL_DATA is defined
            gdal_data_dir = os.environ.get('GDAL_DATA')
    if gdal_data_dir is not None and os.path.exists(gdal_data_dir):
        log.info("Copying GDAL_DATA from %s" % gdal_data_dir)
        copy_data_tree(gdal_data_dir, gdal_data_destination)
    else:
        raise SystemExit("GDAL_DATA directory not found. Unable to package data in wheels.")

    # Conditionally copy PROJ DATA.
    proj_data_dir = os.environ.get('PROJ_DATA', os.environ.get('PROJ_LIB', '/usr/local/share/proj'))
    if os.path.exists(proj_data_dir):
        log.info("Copying PROJ_DATA from %s" % proj_data_dir)
        copy_data_tree(proj_data_dir, 'rasterio/proj_data')
    else:
        raise SystemExit("PROJ_DATA directory not found. Unable to package data in wheels.")
else:
    log.info("Not copying GDAL data to the final package")

compile_time_env = {
    "CTE_GDAL_MAJOR_VERSION": gdal_major_version,
    "CTE_GDAL_MINOR_VERSION": gdal_minor_version,
    "CTE_GDAL_PATCH_VERSION": gdal_patch_version,
}

ext_options = {
    'include_dirs': include_dirs,
    'library_dirs': library_dirs,
    'libraries': libraries,
    'extra_link_args': extra_link_args,
    'define_macros': [],
    'cython_compile_time_env': compile_time_env
}

if not os.name == "nt":
    # These options fail on Windows if using Visual Studio
    ext_options['extra_compile_args'] = ['-Wno-unused-parameter',
                                         '-Wno-unused-function']

# Copy extension options for cpp extension modules.
cpp_ext_options = copy.deepcopy(ext_options)

# Remove -std=c++11 from C extension options.
try:
    ext_options['extra_link_args'].remove('-std=c++11')
    ext_options['extra_compile_args'].remove('-std=c++11')
except Exception:
    pass

cpp11_flag = '-std=c++11'

# 'extra_compile_args' may not be defined
eca = cpp_ext_options.get('extra_compile_args', [])

if platform.system() == 'Darwin':

    if cpp11_flag not in eca:
        eca.append(cpp11_flag)

    eca += [cpp11_flag, '-mmacosx-version-min=10.9', '-stdlib=libc++']

# TODO: Windows

elif cpp11_flag not in eca:
    eca.append(cpp11_flag)

cpp_ext_options['extra_compile_args'] = eca

# Configure optional Cython coverage.
cythonize_options = {"language_level": sys.version_info[0], "compiler_directives": {"freethreading_compatible": True}}
if os.environ.get('CYTHON_COVERAGE'):
    cythonize_options['compiler_directives'].update(linetrace=True)
    cythonize_options['annotate'] = True
    ext_options['define_macros'].extend(
        [('CYTHON_TRACE', '1'), ('CYTHON_TRACE_NOGIL', '1')])

log.debug('ext_options:\n%s', pprint.pformat(ext_options))

ext_modules = None
if "clean" not in sys.argv:
    extensions = [
        Extension("rasterio._base", ["rasterio/_base.pyx"], **ext_options),
        Extension("rasterio._io", ["rasterio/_io.pyx"], **ext_options),
        Extension("rasterio._features", ["rasterio/_features.pyx"], **ext_options),
        Extension("rasterio._env", ["rasterio/_env.pyx"], **ext_options),
        Extension("rasterio._warp", ["rasterio/_warp.pyx"], **cpp_ext_options),
        Extension("rasterio._fill", ["rasterio/_fill.pyx"], **cpp_ext_options),
        Extension("rasterio._err", ["rasterio/_err.pyx"], **ext_options),
        Extension("rasterio._example", ["rasterio/_example.pyx"], **ext_options),
        Extension("rasterio._version", ["rasterio/_version.pyx"], **ext_options),
        Extension("rasterio.cache", ["rasterio/cache.pyx"], **ext_options),
        Extension("rasterio.crs", ["rasterio/crs.pyx"], **ext_options),
        Extension("rasterio.shutil", ["rasterio/shutil.pyx"], **ext_options),
        Extension("rasterio._transform", ["rasterio/_transform.pyx"], **ext_options),
        Extension("rasterio._filepath", ["rasterio/_filepath.pyx"], **cpp_ext_options),
        Extension(
            "rasterio._vsiopener", ["rasterio/_vsiopener.pyx"], **ext_options
        ),
    ]
    ext_modules = cythonize(
        extensions, quiet=True, compile_time_env=compile_time_env, **cythonize_options
    )

setup_args = {}
if os.environ.get('PACKAGE_DATA'):
    setup_args['package_data'] = {'rasterio': ['gdal_data/*', 'proj_data/*']}

# See pyproject.toml for project metadata
setup(
    name="rasterio",  # need by GitHub dependency graph
    ext_modules=ext_modules,
    **setup_args,
)
